use std::sync::Arc;

use sqlx::{migrate::MigrateDatabase, sqlite::SqlitePool, Pool, Sqlite};

#[derive(Debug, sqlx::FromRow)]
pub(crate) struct Device {
    pub(crate) id: i64,
    pub(crate) node_num: u32,
    pub(crate) long_name: String,
    pub(crate) space: String,
    pub(crate) control_room: String,
}

#[derive(Debug, sqlx::FromRow)]
pub(crate) struct RemoteUser {
    pub(crate) id: i64,
    pub(crate) device: i64,
    pub(crate) node_id: u64,
    pub(crate) node_short_name: String,
    pub(crate) node_name: String,
    pub(crate) room: String,
}

#[derive(Debug, sqlx::FromRow)]
struct MatrixState {
    key: String,
    value: String,
}

pub(crate) struct DB {
    pool: Pool<Sqlite>,
}

pub(crate) async fn setup(db_url: String) -> anyhow::Result<Arc<DB>> {
    if !Sqlite::database_exists(&db_url).await.unwrap_or(false) {
        log::debug!("Creating database {}", &db_url);
        match Sqlite::create_database(&db_url).await {
            Ok(_) => log::debug!("Created db successfully"),
            Err(error) => panic!("error: {}", error),
        }
    }

    log::debug!("connecting to database {}", db_url);
    let pool = SqlitePool::connect(&db_url).await?;

    log::info!("running migrations");
    sqlx::migrate!("./migrations")
        .run(&pool)
        .await
        .expect("error running db migrations");

    Ok(Arc::new(DB { pool: pool }))
}

impl DB {
    pub async fn get_device(
        self: &DB,
        node_num: i32,
    ) -> Result<Option<Device>, Box<dyn std::error::Error>> {
        let query = sqlx::query_as::<_, Device>("SELECT * FROM devices WHERE node_num = ? LIMIT 1");
        let device = query.bind(node_num).fetch_one(&self.pool).await?;
        Ok(Option::Some(device))
    }

    pub async fn upsert_device(self: DB, device: Device) -> Result<(), Box<dyn std::error::Error>> {
        let query = sqlx::query("INSERT INTO devices (node_num, space, control_room) VALUES (?, ?, ?) ON CONFLICT(node_num) DO UPDATE SET space=excluded.space, control_room=excluded.control_room) WHERE node_num=excluded.node_num;");

        query
            .bind(device.node_num)
            .bind(device.space)
            .bind(device.control_room)
            .execute(&self.pool)
            .await?;

        Ok(())
    }

    pub async fn get_matrix_session(
        self: &DB,
    ) -> Result<Option<matrix_sdk::matrix_auth::MatrixSession>, Box<dyn std::error::Error>> {
        let session = match self.get_matrix_state("session".to_string()).await? {
            Some(session_json) => serde_json::from_str(&session_json)?,
            None => None,
        };

        Ok(session)
    }

    pub async fn set_matrix_session(
        self: &DB,
        session: matrix_sdk::matrix_auth::MatrixSession,
    ) -> anyhow::Result<()> {
        let session_json = serde_json::to_string(&session)?;

        self.set_matrix_state("session".to_string(), session_json)
            .await
    }

    pub async fn get_matrix_sync_token(self: &DB) -> anyhow::Result<Option<String>> {
        self.get_matrix_state("session_store".to_string()).await
    }

    pub async fn set_matrix_sync_token(self: &DB, token: String) -> anyhow::Result<()> {
        self.set_matrix_state("session_store".to_string(), token)
            .await
    }

    async fn set_matrix_state(self: &DB, key: String, value: String) -> anyhow::Result<()> {
        let query = sqlx::query("INSERT INTO matrix_state (key, value) VALUES (?, ?) ON CONFLICT(key) DO UPDATE SET value=excluded.value");

        query.bind(key).bind(value).execute(&self.pool).await?;

        Ok(())
    }

    pub async fn get_matrix_state(self: &DB, key: String) -> anyhow::Result<Option<String>> {
        let query = sqlx::query_as::<_, MatrixState>(
            "SELECT value FROM matrix_state WHERE 'key' = ? LIMIT 1",
        );
        let result = match query.bind(key).fetch_optional(&self.pool).await.unwrap() {
            Some(row) => Some(row.value),
            None => None,
        };

        Ok(result)
    }

    pub async fn close(self: &DB) {
        self.pool.close().await
    }
}
