use matrix_sdk::ruma::RoomId;
use tokio::signal;
use tokio::sync::mpsc;

mod bridge;
mod config;
mod db;
mod matrix;
mod meshtastic;

#[tokio::main]
async fn main() -> anyhow::Result<()> {
    env_logger::init();

    let config = config::read_config().await;

    let db = db::setup(config.db).await.expect("error connecting to db");

    let (matrix_tx, matrix_rx) = mpsc::channel(32);

    let room_id = RoomId::parse(&config.matrix.room).expect("invalid room id");
    log::debug!("matrix room: {:?}", room_id);

    let matrix_client = matrix::build(config.matrix, &db)
        .await
        .expect("error logging into matrix");

    let meshtastic_client = meshtastic::build(config.meshtastic)
        .await
        .expect("error connecting to meshtastic");

    log::info!("matrix and meshtatic clients ready, spawning threads");

    let mut tasks = vec![];

    tasks.push(tokio::spawn(matrix::sender(
        matrix_client,
        matrix_rx,
        room_id,
    )));
    tasks.push(tokio::spawn(matrix::sync(matrix_client, db)));
    tasks.push(tokio::spawn(meshtastic_client.receive(matrix_tx)));

    match signal::ctrl_c().await {
        Ok(()) => {}
        Err(err) => {
            log::error!("Unable to listen for shutdown signal: {}", err);
        }
    }

    log::info!("shutting down tasks");

    meshtastic_client.close().await;

    futures::future::join_all(tasks).await;

    log::debug!("shutting down database connection");
    db.close().await;
    log::debug!("db connection closed");

    Ok(())
}
