use std::sync::Arc;

use crate::{config, db};
use log;
use matrix_sdk::{
    config::SyncSettings,
    ruma::{
        api::client::filter::FilterDefinition,
        events::room::{
            member::StrippedRoomMemberEvent,
            message::{RoomMessageEventContent, SyncRoomMessageEvent},
        },
        OwnedRoomId, TransactionId, UserId,
    },
    Client, Error, LoopCtrl, Room,
};
use tokio::sync::mpsc::Receiver;
use tokio::time::{sleep, Duration};

pub(crate) async fn build(
    config: config::MatrixConfig,
    db: &db::DB,
) -> Result<matrix_sdk::Client, anyhow::Error> {
    log::debug!("creating matrix client");
    let user = UserId::parse(config.username)
        .expect("invalid matrix username - should be a full mxid with server");

    log::debug!("storing matrix client state in {}", config.state_dir);
    let client = Client::builder()
        .server_name(user.server_name())
        .sqlite_store(config.state_dir, Some(&config.state_passphrase))
        .build()
        .await?;

    log::debug!("checking if there's an existing session");
    match db.get_matrix_session().await.unwrap() {
        Some(session) => {
            log::debug!("restoring session from database");
            client.restore_session(session).await?;
        }
        None => {
            log::debug!("no existing session. Logging in");
            let matrix_auth = client.matrix_auth();
            matrix_auth
                .login_username(user, &config.password)
                .initial_device_display_name(&config.device_name)
                .send()
                .await?;

            log::debug!("creating new session");
            let session = matrix_auth
                .session()
                .expect("A logged-in matrix client should have a session");

            db.set_matrix_session(session).await.unwrap();
        }
    }

    client.add_event_handler(|ev: SyncRoomMessageEvent| async move {
        log::debug!("[matrix] Received a message {:?}", ev);
    });

    client.add_event_handler(on_stripped_state_member);

    log::debug!("connected to matrix, syncing");

    client.sync_once(SyncSettings::default()).await?;

    log::info!("connected to matrix server and synchronized state, ready to continue startup");

    return Ok(client);
}

async fn on_stripped_state_member(
    room_member: StrippedRoomMemberEvent,
    client: Client,
    room: Room,
) {
    if room_member.state_key != client.user_id().unwrap() {
        return;
    }

    tokio::spawn(async move {
        log::debug!("Joining room {}", room.room_id());
        let mut delay = 2;

        while let Err(err) = room.join().await {
            // retry autojoin due to synapse sending invites, before the
            // invited user can join for more information see
            // https://github.com/matrix-org/synapse/issues/4345
            log::error!(
                "Failed to join room {} ({err:?}), retrying in {delay}s",
                room.room_id(),
            );

            sleep(Duration::from_secs(delay)).await;
            delay *= 2;

            if delay > 3600 {
                log::error!("Can't join room {} ({err:?})", room.room_id());
                break;
            }
        }
        log::debug!("Successfully joined room {}", room.room_id());

        let content = RoomMessageEventContent::text_plain("fuck you");
        let txn_id = TransactionId::new();

        let result = room.send(content).with_transaction_id(&txn_id).await;
        match result {
            Ok(_) => log::error!("sent hello message"),
            Err(err) => log::error!(
                "Error sending hello message to room {}: {err:?}",
                room.room_id()
            ),
        }
    });
}

pub(crate) async fn sender(
    client: Client,
    mut rx: Receiver<RoomMessageEventContent>,
    room_id: OwnedRoomId,
) -> anyhow::Result<()> {
    let room = match client.get_room(&room_id) {
        Some(room) => room,
        None => panic!("requested matrix room not found"),
    };

    while let Some(msg) = rx.recv().await {
        match room.send(msg).await {
            Err(err) => log::error!("error sending message to matrix: {:?}", err),
            Ok(response) => log::debug!("sent message to matrix: {:?}", response),
        }
    }

    Ok(())
}

pub(crate) async fn sync(client: Client, db: Arc<db::DB>) -> anyhow::Result<()> {
    let filter = FilterDefinition::with_lazy_loading();
    let mut sync_settings = SyncSettings::default().filter(filter.into());

    if let Some(token) = db
        .get_matrix_sync_token()
        .await
        .map_err(|err| Error::UnknownError(err.into()))?
    {
        sync_settings = sync_settings.token(token);
    }

    client
        .sync(SyncSettings::default())
        .await
        .expect("error syncing matrix");

    loop {
        match client.sync_once(sync_settings.clone()).await {
            Ok(response) => {
                // This is the last time we need to provide this token, the sync method after
                // will handle it on its own.
                sync_settings = sync_settings.token(response.next_batch.clone());
                db.set_matrix_sync_token(response.next_batch).await?;
                break;
            }
            Err(error) => {
                log::error!("An error occurred during initial sync: {error}");
                log::error!("Trying again…");
            }
        }
    }

    log::info!("The client is ready! Listening to new messages…");

    // Now that we've synced, let's attach a handler for incoming room messages.
    // client.add_event_handler(on_room_message);

    // This loops until we kill the program or an error happens.
    client
        .sync_with_result_callback(sync_settings, |sync_result| async move {
            let response = sync_result?;

            // We persist the token each time to be able to restore our session
            db.set_matrix_sync_token(response.next_batch)
                .await
                .map_err(|err| Error::UnknownError(err.into()))?;

            Ok(LoopCtrl::Continue)
        })
        .await?;

    Ok(())
}
